﻿/* Copyright (C) 2007-2010 STS Soft

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace STSdb.General.Streams
{
    /// <summary>
    /// An optimized FileStram - optimizes calls to Seek & Size methods
    /// The requirement is if the file is opened for writing, it is an exclusive.
    /// </summary>
    public class OptimizedFileStream : FileStream
    {
        private long length = long.MinValue;

        public OptimizedFileStream(string fileName, FileMode mode)
            : base(fileName, mode)
        {
        }

        public OptimizedFileStream(string fileName, FileMode mode,FileAccess access,FileShare fileShare)
            : base(fileName, mode, access, fileShare)
        {
        }

        public override long Position
        {
            get { return base.Position; }
            set
            {
                if (base.Position != value)
                    base.Position = value;
            }
        }

        public override void Write(byte[] array, int offset, int count)
        {
            try
            {
                base.Write(array, offset, count);

                if (Position > Length)
                    length = Position;
            }
            catch (Exception exc)
            {
                length = long.MinValue;
                throw exc;
            }
        }

        public override void WriteByte(byte value)
        {
            Write(new byte[] { value }, 0, 1);
        }

        public override long Length
        {
            get
            {
                if (length == long.MinValue)
                    length = base.Length;

                return length;
            }
        }

        public override void SetLength(long value)
        {
            try
            {
                base.SetLength(value);

                length = value;
            }
            catch (Exception exc)
            {
                length = long.MinValue;

                throw exc;
            }
        }

        public override long Seek(long offset, SeekOrigin origin)
        {
            switch (origin)
            {
                case SeekOrigin.Begin:
                    {
                        if (offset != Position)
                            return base.Seek(offset, origin);
                    }
                    break;
                case SeekOrigin.Current:
                    {
                        if (offset != 0)
                            return base.Seek(offset, origin);
                    }
                    break;
                case SeekOrigin.End:
                    {
                        if (offset != Length - Position)
                            return base.Seek(offset, origin);
                    }
                    break;
            }

            return Position;
        }
    }
}
